#Copyright Massachusetts Institute of technology, 1998.
#Written by Eric Prud'hommeaux for the World Wide Web Consortium

use strict;
require Exporter;

use W3C::Database::DBIInterface;
use W3C::Database::MysqlObjects;

$W3C::Rdf::MysqlDB::REVISION = '$Id: MysqlDB.pm,v 1.20 1999/12/14 13:14:40 eric Exp $ ';

package W3C::Rdf::MysqlDB;
use vars qw($VERSION $DSLI @ISA @EXPORT @EXPORT_OK
	    $NodeIdType_Fake $NodeIdType_Gen $NodeIdType_Ref $NodeIdType_String
	    $FixedId_Fake $FixedID_Gen $FixedID_Ref $FixedID_String
	    $ContainerIdType_Seq $ContainerIdType_Bag $ContainerIdType_Alt
	    $AttributionType_SOURCE $AttributionType_GEN $AttributionType_REI);
@ISA = qw(Exporter);
@EXPORT = qw($NodeIdType_Fake $NodeIdType_Gen $NodeIdType_Ref $NodeIdType_String
	     $ContainerIdType_Seq $ContainerIdType_Bag $ContainerIdType_Alt
	     $AttributionType_SOURCE $AttributionType_GEN $AttributionType_REI);
$VERSION = 0.94;
$DSLI = 'adpO';
#($NodeIdType_ID, $NodeIdType_Ref, $NodeIdType_Gen, $NodeIdType_Fake, $NodeIdType_String) = 
#    ('ID', 'Ref', 'Gen', 'Fake', 'String');
($NodeIdType_Fake, $NodeIdType_Gen, $NodeIdType_Ref, $NodeIdType_String) = 
    ('Fake', 'Gen', 'Ref', 'String');
($FixedId_Fake, $FixedID_Gen, $FixedID_Ref, $FixedID_String) = 
    (9, 12, 11, 10);
($ContainerIdType_Seq, $ContainerIdType_Bag, $ContainerIdType_Alt) = 
    ('Seq', 'Bag', 'Alt');
($AttributionType_SOURCE, $AttributionType_GEN, $AttributionType_REI) = 
    ('Source', 'Generated', 'Reified');


######### ObjectDB Interface #########

# import defs generated by W3C::Database::bin::object_maker
use W3C::Rdf::RdfObjects;
@W3C::Rdf::RdfObjects::Containers::EXPORTS_OK = qw(%string_to_containerType);

#### base class for the RdfObjects ####
package W3C::Rdf::MysqlElement;
use W3C::Database::ObjectBase;
@W3C::Rdf::MysqlElement::ISA = qw(W3C::Database::ObjectBase);

sub show1 ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);
    my $ret = '';
    return $self;
}

sub show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my $ret = $self->showSelf($flags).'- bogus ref:'.$self->{TABLE_NAME}.':'.$self->{DB_ID}.' -';
    return $ret;
}

sub showSelf ($) {
    my ($self, $flags) = @_;
    my $ret = '';
    if ($flags->{-showRef}) {
	$ret .= $self.' ';
    } elsif ($flags->{-showTypes}) {
	$ret .= (ref $self).' ';
    }
    $ret .= $self->{DB_ID}.' ' if ($flags->{-showIds});
    $ret .= "\n" if ($flags->{-lineBreaks});
    return $ret;
}

#### show and getUri methods for all the RdfObjects ####

# push (@W3C::Rdf::RdfObjects::Uris::ISA, qw(URI::URL)); # may as well provide the URI interface - someday

sub W3C::Rdf::RdfObjects::Uris::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($uri) = 
	($self->_getUri);
    if ($namespaceHandler) {
	$uri = $namespaceHandler->unmapNamespace($uri);
    }
    my $ret = $self->showSelf($flags).$uri;
    return $ret;
}

sub W3C::Rdf::RdfObjects::Fragments::getUri () {
    my ($self) = @_;
    my $ret = $self->_getUri->_getUri;
    my $fragment = $self->_getFragment;
    $ret .= '#'.$fragment if ($fragment);
    return $ret;
}

sub W3C::Rdf::RdfObjects::Fragments::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($uri, $fragment) = 
	($self->_getUri, $self->_getFragment);
    my $ret = $self->showSelf($flags).$uri->show($prefix, undef, $flags);
    $ret .= '#'.$fragment if ($fragment);
    if ($namespaceHandler) {
	$ret = $namespaceHandler->unmapNamespace($ret);
    }
    return $ret;
}

sub W3C::Rdf::RdfObjects::Attributions::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);
    return '' if (!$flags->{-showAttributions});
    my ($type, $uri, $parent, $source, $trusted) = 
	($self->_getType, $self->_getUri, $self->_getParent, $self->_getSource, $self->_getTrusted);
    my $ret = $self->showSelf($flags);
    $ret .= $trusted ? '+' : 'x' if ($flags->{-showTrust});
    my %typeStrings = ('Source' => 's', 'Generated' => 'g', 'Reified' => 'r');
    $ret .= $typeStrings{$type}.' ' if ($flags->{-showType});
    $ret .= $uri->show($prefix, $namespaceHandler, $flags);
    $ret .= ' '.$parent if ($flags->{-showParent});
    return $ret;
}

sub W3C::Rdf::RdfObjects::Attributions::getSource () {
    return $_[0]->_getUri;
}

sub W3C::Rdf::RdfObjects::Attributions::getUri () {
    my ($self) = @_;
    return $self->_getUri->_getUri->_getUri;
}

sub W3C::Rdf::RdfObjects::Strings::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($string) = 
	($self->_getString);
    my $ret = $self->showSelf($flags).'"'.$string.'"';
    return $ret;
}

sub W3C::Rdf::RdfObjects::GenIds::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($attribution) = 
	($self->_getAttribution);
#    my $ret = $self->showSelf($flags).$attribution->show($prefix, $namespaceHandler, $flags).'#genId_'.$dbId;
    my $uri = $attribution->getUri;
    $uri = $namespaceHandler->unmapNamespace($uri) if ($namespaceHandler);
    my $ret = $self->showSelf($flags).$uri.'#genId_'.$self->_getId;
    return $ret;
}

sub W3C::Rdf::RdfObjects::RdfIds::getUri () {
    my ($self) = @_;
    return $self->_getUri ? $self->_getUri->getUri : $self->_getGenId ? ('genId'.$self->_getGenId->getId) : undef;
}

sub W3C::Rdf::RdfObjects::RdfIds::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($type, $uri, $genId, $string) = 
	($self->_getType, $self->_getUri, $self->_getGenId, $self->_getString);
    my $value;
    $type = $W3C::Rdf::RdfObjects::RdfIds::string_to_type{$type} if ($type !~ m/^\d+$/);
    if ($type == $W3C::Rdf::RdfObjects::RdfIds::type_GEN) {
	$value = $genId;
    } elsif ($type == $W3C::Rdf::RdfObjects::RdfIds::type_REF) {
	$value = $uri;
    } elsif ($type == $W3C::Rdf::RdfObjects::RdfIds::type_STRING) {
	$value = $string;
    }
    my $ret = $self->showSelf($flags);
    $ret .= $type.' ' if ($flags->{-showType});
    eval {$value->show;}; if ($@) {
	$ret = 'bogus '.(ref $self).':'.$self->_getId;
    } else {
	$ret .= $value->show($prefix, $namespaceHandler, $flags);
    }
    return $ret;
}

sub W3C::Rdf::RdfObjects::Descriptions::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($type, $about, $aboutEach, $aboutEachPrefix, $attribution) = 
	($self->_getType, $self->_getAbout, $self->_getAboutEach, $self->_getAboutEachPrefix, $self->_getAttribution);
    my $ret = $self->showSelf($flags);
    $ret .= ' '.$type->show if ($type);
    $ret .= ' about:'.$about->show($prefix, $namespaceHandler, $flags) if ($about);
    $ret .= ' aboutEach:'.$aboutEach->show($prefix, $namespaceHandler, $flags) if ($aboutEach);
    $ret .= ' aboutEachPrefix:'.$aboutEachPrefix->show($prefix, $namespaceHandler, $flags) if ($aboutEachPrefix);
    $ret .= ' '.$attribution->show($prefix, $namespaceHandler, $flags);
    return $ret;
}

# inherit from W3C::Rdf::RdfDB::MappedNode to inherit addStatement functionality
push (@W3C::Rdf::RdfObjects::MappedNodes::ISA, qw(W3C::Rdf::RdfDB::MappedNode));

sub W3C::Rdf::RdfObjects::MappedNodes::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($rdfId, $mapFrom, $description, $container, $attribution) = 
	($self->_getRdfId, $self->_getMapFrom, $self->_getDescription, $self->_getContainer, $self->_getAttribution);
    my $ret = '('.$self->showSelf($flags);
    $ret .= $mapFrom->show($prefix, $namespaceHandler, $flags).' ';
    $ret .= $rdfId->show($prefix, $namespaceHandler, $flags).' *';
    $ret .= ' for '.$description->show($prefix, $namespaceHandler, $flags) if ($flags->{-showFors} && defined $description);
    $ret .= $attribution->show($prefix, $namespaceHandler, $flags); $ret .= ')';
    return $ret;
}

sub W3C::Rdf::RdfObjects::MappedNodes::mapToTriple ($) {
    my ($self, $triple) = @_;
    if ($triple->_getPredicate == $self->_getMapFrom && $triple->getSubject == $self->_getRdfId) {
	return $triple->getObject;
    } else {
	return $self;
    }
}

sub W3C::Rdf::RdfObjects::Containers::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($rdfId, $type, $nextOrdinal, $attribution) = 
	($self->_getRdfId, $self->_getType, $self->_getNextOrdinal, $self->_getAttribution);
    my $ret = $self->showSelf($flags).$rdfId->show($prefix, $namespaceHandler, $flags).' '.$type.' '.$nextOrdinal.' '.$attribution->show($prefix, $namespaceHandler, $flags);
    return $ret;
}

# sub W3C::Rdf::RdfObjects::Containers::addElement ($$$$) {
#     my ($self, $revision, $rdfIdOrOrdinal, $description, $attribution) = @_;
#     my $rdfId;
#     if ($rdfIdOrOrdinal->isa(qw(W3C::Rdf::RdfObjects::RdfIds))) {
# 	$rdfId = $rdfIdOrOrdinal;
#     } else {
# 	my $name = $self->{RDF_DB}->{RDF_SCHEMA_URI}.'_'.$rdfIdOrOrdinal;
# 	my $elementFragment = $self->{RDF_DB}->_getFragment($revision, $name);
# 	$rdfId = $self->{RDF_DB}->_getRdfId($NodeIdType_Ref, $elementFragment);
#     }
#     my $newNode = new W3C::Rdf::MysqlNode($self->{RDF_DB}, $rdfId, $description, $self, $attribution);
#     die $self->{RDF_DB}."->addNode($newNode)";
# }

sub W3C::Rdf::RdfObjects::Statements::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($predicate, $subject, $object, $container, $reification, $attribution) = 
	($self->_getPredicate, $self->_getSubject, $self->_getObject, $self->_getContainerId, $self->_getReifiedId, $self->_getAttribution);
    my $ret;
    $ret .= $self->showSelf($flags);
    $ret .= $attribution->show($prefix, $namespaceHandler, {-showAttributions => 1, -showTrust => 1, -showType => 1, %$flags});
    $ret .= ' triple('.$predicate->show($prefix, $namespaceHandler, $flags).',';
    $ret .= $subject->show($prefix, $namespaceHandler, $flags).',';
    $ret .= $object->show($prefix, $namespaceHandler, $flags).')';
    if (!$flags->{-noReifications}) {
	$ret .= ' reified as '.$reification->show($prefix, $namespaceHandler, $flags) if ($reification);
	$ret .= ' in '.$container->show($prefix, $namespaceHandler, $flags) if ($container);
    }
    return $ret;
}

sub W3C::Rdf::RdfObjects::Statements::getPredicate () {return shift->_getPredicate}
sub W3C::Rdf::RdfObjects::Statements::getSubject () {return shift->_getSubject}
sub W3C::Rdf::RdfObjects::Statements::getObject () {return shift->_getObject}
sub W3C::Rdf::RdfObjects::Statements::getAttribution () {return shift->_getAttribution}

# inherit from W3C::Rdf::RdfDB::MappedStatement to inherit addStatement functionality
push (@W3C::Rdf::RdfObjects::MappedStatements::ISA, qw(W3C::Rdf::RdfDB::MappedStatement));

sub W3C::Rdf::RdfObjects::MappedStatements::show ($;$$\%) {
    my ($self, $prefix, $namespaceHandler, $flags) = @_; $prefix = '  ' if (!defined $prefix); $flags = {} if (!defined $flags);

    my ($predicate, $subject, $subjectMap, $object, $objectMap, $container, $reification, $attribution) = 
	($self->_getPredicate, $self->_getSubject, $self->_getSubjectMap, $self->_getObject, $self->_getObjectMap, $self->_getContainerId, $self->_getReifiedId, $self->_getAttribution);
    my $ret = $self->showSelf($flags).'('.$predicate->show($prefix, $namespaceHandler, $flags).',';
    eval {
	$ret .= $subject ? $subject->show($prefix, $namespaceHandler, $flags) : 
	    $subjectMap->show($prefix, $namespaceHandler, $flags);
	$ret .= ',';
	$ret .= $object ? $object->show($prefix, $namespaceHandler, $flags) : 
	    $objectMap->show($prefix, $namespaceHandler, $flags);
	$ret .= ')';
	$ret .= ' reified as '.$reification->show($prefix, $namespaceHandler, $flags);
	$ret .= ' in '.$container->show($prefix, $namespaceHandler, $flags) if ($container);
	$ret .= ' by '.$attribution->show($prefix, $namespaceHandler, {-showAttributions => 1, %$flags});
    };
    if ($@) {
	warn $@;
	print $ret."\n";
    }
    return $ret;
}

# don't overload W3C::RdfDB::MappedStatement::mapToTriple because the generated data need not go to database
sub W3C::Rdf::RdfObjects::MappedStatements::getPredicate () {
    return $_[0]->_getPredicate ? $_[0]->_getPredicate : $_[0]->_getPredicateMap}
sub W3C::Rdf::RdfObjects::MappedStatements::getSubject () {
    return $_[0]->_getSubject ? $_[0]->_getSubject : $_[0]->_getSubjectMap}
sub W3C::Rdf::RdfObjects::MappedStatements::getObject () {
    return $_[0]->_getObject ? $_[0]->_getObject : $_[0]->_getObjectMap}
sub W3C::Rdf::RdfObjects::MappedStatements::getContainer () {
    return $_[0]->_getContainerId}
sub W3C::Rdf::RdfObjects::MappedStatements::getReification () {
    return $_[0]->_getReifiedId}
sub W3C::Rdf::RdfObjects::MappedStatements::getAttribution () {
    return $_[0]->_getAttribution}


######### quick and dirty W3C::Rdf::RdfDBException #########

package W3C::Rdf::RdfDBException;
use W3C::SAX::SAXException;
@W3C::Rdf::RdfDBException::ISA = ('W3C::SAX::SAXException');


######### the long-awaited MysqlDB #########

use W3C::Rdf::RdfDB;
package W3C::Rdf::MysqlDB;
@W3C::Rdf::MysqlDB::ISA = qw(W3C::Rdf::RdfDB);

sub new {
    my ($proto, $parms) = @_;
    my $class = ref($proto) || $proto;
    my $self = $class->SUPER::new($parms);
    bless ($self, $class);
    $self->{OB} = new W3C::Database::MysqlObjects($parms->{-properties}, \%W3C::Rdf::RdfObjects::_AllTables, 
						  \@W3C::Rdf::RdfObjects::_TableOrder);
    $self->setRdfPrefix('http://www.w3.org/1999/02/22-rdf-syntax-ns#');
    return $self;
}

sub setErrorHandler {
    my ($self, $errorHandler) = @_;
    $self->{ERROR_HANDLER} = $errorHandler;
}

sub setSource {
    my ($self, $source) = @_;
    $self->{SOURCE} = $source;
}

# get and set the RDF schema URI
sub setRdfPrefix {
    my ($self, $newUri) = @_;
    $self->{RDF_SCHEMA_URI} = $newUri;
    $self->{SCHEMA_URI_OBJECT} = $self->getUri($newUri);
    $self->{RDFID_RDF_LI} = $self->getUri('#li', $self->{SCHEMA_URI_OBJECT});
    $self->{RDFID_RDF_DESCRIPTION} = $self->getUri('#Description', $self->{SCHEMA_URI_OBJECT});
}

sub getRdfPrefix {
    my ($self) = @_;
    return $self->{RDF_SCHEMA_URI};
}

# attribution interface
sub setSourceAttribution {
    my ($self, $sourceAttribution) = @_;
    $self->{CURRENT_SOURCE_ATTRIBUTION} = $sourceAttribution;
    $self->{CURRENT_GENERATED_ATTRIBUTION} = $self->assureGeneratedAttribution($sourceAttribution);
}

sub getSourceAttribution {
    my ($self) = @_;
    return $self->{CURRENT_SOURCE_ATTRIBUTION};
}

sub makeSourceAttribution {
    my ($self, $trusted, $source, $parent) = @_;
    return new W3C::Rdf::SourceAttribution($self, $trusted, $source, $parent);
}

my %GeneratedAttributions;

sub assureGeneratedAttribution {
    my ($self, $sourceAttribution) = @_;
    my $ret = $GeneratedAttributions{$sourceAttribution};
    if (!defined $ret) {
	$ret = $self->getAttribution($AttributionType_GEN, $sourceAttribution->_getUri, $sourceAttribution, 1, undef);
	$GeneratedAttributions{$sourceAttribution} = $ret;
    }
    return $ret;
}

# namespace support - still in flux: not sure whether namespaces will
# be expanded by downcalls from the XML parser or by a W3C::SAX::HandlerStream.
sub getNamespace {
    my ($self, $sAs) = @_;
    my $lower = $sAs;
    $lower =~ tr/A-Z/a-z/;
    return $self->{NAMESPACES}->{$lower};
}

sub mapNamespace {
    my ($self, $predicate) = @_;
    return $predicate; # !!!
    my ($namespace, $property) = $predicate =~ m/(?:([^:]*):)?(.*)/;
    return $self->{NAMESPACES}{$namespace}.$property;
}


######### RdfDB Object Creation Interface #########

sub getUri ($$) {
    my ($self, $uriParm, $base) = @_;
    $base = $base->getUri if ($base);
    my $uriOb = new URI::URL($uriParm, $base)->abs;
    my $uri = $self->{OB}->ensure('Uris', 'u_uri', {'uri' => $uriOb->scheme.':'.$uriOb->opaque}, {-replace => 1});
    my $fragmentStr = $uriOb->fragment;
    my $fragment = $self->{OB}->ensure('Fragments', 'u_uri_fragment', {'uri' => $uri, 'fragment' => $fragmentStr}, 
				       {-replace => 1, -default => {'fragment' => 1}});
    return $self->getRdfId($NodeIdType_Ref, $fragment);
}

sub getAttribution ($$$$$$) {
    my ($self, $type, $fragmentId, $parent, $source, $trusted, $clear) = @_;
    die if (!$fragmentId->isa(qw(W3C::Rdf::RdfObjects::RdfIds)) && !$fragmentId->isa(qw(W3C::Rdf::RdfObjects::Fragments)));
#    my $fragment = $fragmentId->isa(qw(W3C::Rdf::RdfObjects::Fragments)) ? $fragmentId : 
#	$self->{OB}->ensure('Fragments', 'u_uri_fragment', {'uri'=>$fragmentId, 'fragment'=>''}, {-replace => 1});
    my $fragment = $fragmentId->isa(qw(W3C::Rdf::RdfObjects::Fragments)) ? $fragmentId : $fragmentId->_getUri;
    my ($created, $attrib) = $self->{OB}->ensure('Attributions', 'u_type_uri_parent', 
						 {'type' => $type, 'uri' => $fragment, 'parent' => $parent, 
						  'source' => $source, 'trusted' => $trusted}, 
						 {-replace => 1, -default => {'parent' => 1}});
    my $id = $attrib->getPrimaryKey;
#    print $attrib->show(undef, undef, {-showAttributions => 1, -showRef => 1, -showTypes => 1, -lineBreaks => 1});
    if ($clear) {
	$self->{OB}->executeUpdate('DELETE FROM Statements WHERE attribution='.$id);
	$self->{OB}->executeUpdate('DELETE FROM MappedStatements WHERE attribution='.$id);
	$self->{OB}->executeUpdate('DELETE FROM Descriptions WHERE attribution='.$id);
	$self->{OB}->executeUpdate('DELETE FROM Containers WHERE attribution='.$id);
	$self->{OB}->executeUpdate('DELETE FROM MappedNodes WHERE attribution='.$id);
	$self->{OB}->executeUpdate('DELETE FROM GenIds WHERE attribution='.$id);
	$self->{OB}->executeUpdate('DELETE FROM TypeClosures WHERE attribution='.$id);
#	$self->{OB}->executeUpdate('DELETE FROM AttributionClosures WHERE attribution='.$id); @@@ clean out all other attribs and regen
    }

    if ($created) {
	$self->{OB}->executeUpdate('INSERT INTO AttributionClosures set sub='.$id.',super='.$id.',g=0,attribution='.$id);
	if ($parent) {
	    my @parentClosures;
	    $self->{OB}->executeArrayQuery(\@parentClosures, 'SELECT '.$id.',super,g+1,'.$id.
						' FROM AttributionClosures WHERE sub='.$parent->_getId);
	    foreach my $row (@parentClosures) {
		$self->{OB}->executeUpdate('INSERT INTO AttributionClosures (sub,super,g,attribution) VALUES ('.
						join (',', @$row).')');
	    }
	}
    }
    return $attrib;
}

sub getString ($$) {
    my ($self, $stringParm) = @_;
    my $string = $self->{OB}->ensure('Strings', 'u_substring', 
				     {'substring' => $stringParm, 'string' => $stringParm}, {-replace => 0});
    my $type = $W3C::Rdf::RdfObjects::RdfIds::type_to_string{$W3C::Rdf::RdfObjects::RdfIds::type_STRING};
    return $self->{OB}->ensure('RdfIds', 'u_uri_genId_string', {'type' => $type, 'string' => $string},
			       {-replace => 1, -default => {'uri' => 1, 'genId'=> 1 }});
}

sub createGenId ($) {
    my ($self, $attribution) = @_;
    my $genId = $self->{OB}->add('GenIds', {'attribution' => $attribution}, {-replace => 1});
    my $type = $W3C::Rdf::RdfObjects::RdfIds::type_to_string{$W3C::Rdf::RdfObjects::RdfIds::type_GEN};
    return $self->{OB}->ensure('RdfIds', 'u_uri_genId_string', {'type' => $type, 'genId' => $genId}, 
			       {-replace => 1, -default => {'uri' => 1, 'string' => 1}});
}

sub getRdfId ($$$$) {
    my ($self, $type, $value) = @_;
    my ($uri, $genId, $string);
    if ($type eq $NodeIdType_Gen) {
	($type, $uri, $genId, $string) = ($W3C::Rdf::RdfObjects::RdfIds::type_GEN, undef, $value, undef);
    } elsif ($type eq $NodeIdType_Ref) {
	($type, $uri, $genId, $string) = ($W3C::Rdf::RdfObjects::RdfIds::type_REF, $value, undef, undef);
    } elsif ($type eq $NodeIdType_String) {
	($type, $uri, $genId, $string) = ($W3C::Rdf::RdfObjects::RdfIds::type_STRING, undef, undef, $value);
    }
    my $ret = $self->{OB}->ensure('RdfIds', 'u_uri_genId_string', 
				  {'type' => $type, 'uri' => $uri, 'genId' => $genId, 'string' => $string}, 
				  {-replace => 1, -default => {'uri' => 1, 'genId' => 1, 'string' => 1}});
    return $ret;
}

sub createDescription ($$$) {
    my ($self, $type, $about, $aboutEach, $aboutEachPrefix, $reificationContainer, $attribution) = @_;
    my $desc = $self->{OB}->add('Descriptions', {'type' => $type, 'about' => $about, 'aboutEach' => $aboutEach, 
						 'aboutEachPrefix' => $aboutEachPrefix, 'attribution' => $attribution}, 
				{-replace => 1, -default => {'about' => 1, 'aboutEach' => 1, 'aboutEachPrefix' => 1}});
    my $node = undef;
    if (defined $about) {
	$node = $about;
	my $tclosure = $self->{OB}->ensure('TypeClosures', 'u_node_type_g_attribution', 
					   {'node' => $about, 'type' => $type, 'g' => 0, 'attribution' => $attribution}, 
					   {-replace => 0}) if ($type && $type != $self->{RDFID_RDF_DESCRIPTION});
    } elsif (defined $aboutEach) {
	$node = $self->{OB}->add('MappedNodes',{'rdfId' => $aboutEach, 'mapFrom' => $self->{RDFID_RDF_LI}, 'description' => $desc, 
						'container' => $reificationContainer, 'attribution'=>$attribution, 'type'=>$type}, 
				 {-replace => 0});
    } elsif (defined $aboutEachPrefix) {
	die "!!! need to handle aboutEachPrefix";
    } else {
	$node = $self->createGenId($attribution);
    }
    return ($desc, $node);
}

sub createContainer ($$$) {
    my ($self, $rdfId, $type, $nextOrdinal, $attribution) = @_;
    my ($ContainerIdType_Seq, $ContainerIdType_Bag, $ContainerIdType_Alt) = 
	('Seq', 'Bag', 'Alt');
    my %externalInterfaceToDBEnum = ($ContainerIdType_Bag => $W3C::Rdf::RdfObjects::Containers::type_BAG, 
				     $ContainerIdType_Seq => $W3C::Rdf::RdfObjects::Containers::type_SEQ, 
				     $ContainerIdType_Alt => $W3C::Rdf::RdfObjects::Containers::type_ALT);
    my $containerType = $externalInterfaceToDBEnum{$type};
    die "Bad container type: \"$type\"." if (!defined $containerType);
    return $self->{OB}->add('Containers', {'rdfId' => $rdfId, 'type' => $containerType, 
					   'nextOrdinal' => $nextOrdinal, 'attribution' => $attribution}, {-replace => 1});
}


######### Overloaded RdfDB addStatement subfunctions #########

sub hasMappedNodes ($$$) {
    my ($self, $predicate, $subject, $object) = @_;
    my ($mappedPredicate, $mappedSubject, $mappedObject) = 
	($predicate->isa(qw(W3C::Rdf::RdfObjects::MappedNodes)), 
	 $subject->isa(qw(W3C::Rdf::RdfObjects::MappedNodes)), 
	 $object->isa(qw(W3C::Rdf::RdfObjects::MappedNodes)));
    return $mappedPredicate || $mappedSubject || $mappedObject;
}

sub mappedStatementsMatchingTriple ($) {
    my ($self, $triple) = @_;
    my $query = 'SELECT MappedStatements.id FROM MappedStatements,MappedNodes WHERE (MappedStatements.subjectMap=MappedNodes.id OR MappedStatements.objectMap=MappedNodes.id) AND MappedNodes.rdfId='.$triple->getSubject->_getId.' AND MappedNodes.mapFrom='.$triple->getPredicate->_getId;
    return $self->loadTable($query, 'MappedStatements');
}

sub triplesMatchingMappedNode ($) {
    my ($self, $potentialMappedNode) = @_;
    if ($potentialMappedNode->isa('W3C::Rdf::RdfObjects::MappedNodes')) {
	my $query;
	if (1) {
	    $query = 'SELECT Statements.object FROM Statements WHERE Statements.subject='.$potentialMappedNode->_getRdfId->_getId.' AND Statements.predicate='.$potentialMappedNode->_getMapFrom->_getId;
	} elsif (0) {
	    $query = 'SELECT Statements.subject FROM Statements WHERE Statements. object='. $potentialMappedNode->object->_getRdfId->_getId.' AND Statements.predicate='. $potentialMappedNode->object->_getMapFrom->_getId;
	} else {
	    $query = 'SELECT Statements.predicate FROM Statements WHERE Statements. object='. $potentialMappedNode->object->_getRdfId->_getId.' AND Statements.subject='. $potentialMappedNode->object->_getMapFrom->_getId;
	}
	return $self->loadTable($query, 'RdfIds');
    } else {
	return ($potentialMappedNode);
    }
}

sub createStatement ($$$$$$) {
    my ($self, $predicate, $subject, $object, $container, $reification, $attribution) = @_;
    my $triple = $self->{OB}->add('Statements', {'predicate' => $predicate, 'subject' => $subject, 'object' => $object, 
						 'containerId'=>$container,'reifiedId'=>$reification,'attribution'=>$attribution}, 
				  {-replace => 1});
    return ([$triple], $triple);
}

sub createMappedStatement ($$$$$$) {
    my ($self, $predicate, $subject, $object, $container, $reification, $attribution) = @_;
    my ($mappedSubject, $mappedObject) = 
	($subject->isa(qw(W3C::Rdf::RdfObjects::MappedNodes)), $object->isa(qw(W3C::Rdf::RdfObjects::MappedNodes)));
    my $mappedStatement = $self->{OB}->add('MappedStatements', {'predicate' => $predicate, 
								$mappedSubject ? 'subjectMap' : 'subject' => $subject, 
								$mappedObject ? 'objectMap' : 'object' => $object, 
								'containerId' => $container, 
								'reifiedId' => $reification, 'attribution' => $attribution}, 
					   {-replace => 1});
    return $mappedStatement;
}

######### end overload the addStatement functions #########


sub reallyReifyNUKE2 {
    my ($self, $triple) = @_;
    my ($nodeId, $container) = $triple->getReificationInfo;
    return () if (!$nodeId);
    my $reificationAttribution = new W3C::Rdf::ReifiedAttribution($triple, $self, 1, undef, $triple->getAttribution);
    my @ret;
#    if (exists $self->{REIFICATIONS}{$triple}) {
#	$self->{ERROR_HANDLER}->error(new W3C::Rdf::MysqlDBException('re-reified triple '.$triple->show.' in '.$self));
#    }
#    $self->{REIFICATIONS}{$triple} = $triple;
    push (@ret, $self->addTriple(new W3C::Rdf::PredicateTriple($self, $nodeId, $triple->_getPredicate, $reificationAttribution)));
    push (@ret, $self->addTriple(new W3C::Rdf::SubjectTriple($self, $nodeId, $triple->_getSubject, $reificationAttribution)));
    push (@ret, $self->addTriple(new W3C::Rdf::ObjectTriple($self, $nodeId, $triple->_getObject, $reificationAttribution)));
    push (@ret, $self->addTriple(new W3C::Rdf::TypeStatementTriple($self, $nodeId, $reificationAttribution)));
    push (@ret, $container->addOrdinalTriple($nodeId, undef, $reificationAttribution)) if ($container);
    return @ret;
}

sub setLazyReificationNUKE2 {
    my ($self, $newState) = @_;
    return () if ($newState == $self->{LAZY_REIFICATION});
    $self->{LAZY_REIFICATION} = $newState;
    my @ret;
    if ($newState) {
	my %doomed;
	foreach my $triple (@{$self->{TRIPLES}}) {
	    $doomed{$triple} = $triple if ($triple->getAttribution->isa('W3C::Rdf::ReifiedAttribution'));
	}
	$self->deleteTriple(\%doomed);
    } else {
	foreach my $triple (@{$self->{TRIPLES}}) {
	    push (@ret, $self->reallyReify($triple));
	}
    }
    return @ret;
}

##### triplet database callback #####

# This slows the addTriple method down so it should only be used
# where needed because the parsing needs to change as a result of
# added triples, as in the case of the trust mechanism in W3C::Rdf::RdfParser.
sub registerTripleObserver {
    my ($self, $triples, $observer) = @_;
    push (@{$self->{OBSERVERS}}, [$triples, $observer]);
}

######### QUERY INTERFACE #########

#####  loadTable - support for sql-specific query engines #####
sub loadTable ($$) {
    my ($self, $query, $table) = @_;
    my @objects;
    my $count = $self->{OB}->executeQuery(\@objects, $query);
    $self->{OB}->load($table, \@objects, \@W3C::Rdf::RdfObjects::_TableOrder);
    my @ret;
    foreach my $id (@objects) {
	push (@ret, $self->{OB}->getById($table, $id));
    }
    return @ret;
}

sub loadArrayTable ($$) {
    my ($self, $query, $tables) = @_;
    my @rows;
    my $count = $self->{OB}->executeArrayQuery(\@rows, $query);

    my @ret;
    foreach my $row (@rows) {
	my @objects = @$row;
	my $startIndex = 0;
	my @row;
	for (my $i = 0; $i < @$tables; $i++) {
	    my ($elements, $table) = @{$tables->[$i]};
	    $self->{OB}->load($table, [@objects[$startIndex..$startIndex+$elements-1]], \@W3C::Rdf::RdfObjects::_TableOrder);
	    foreach my $id (@objects[$startIndex..$startIndex+$elements-1]) {
		push (@row, $self->{OB}->getById($table, $id));
	    }
	    $startIndex += $elements;
	}
	push (@ret, \@row);
    }
    return @ret;
}

#####
# supplied - strain view for supplied triples (ones that were added to DB rather than reified or something like that

sub supplied {
    my ($self, $view) = @_;
    $view ||= $self->{TRIPLES};
    my @ret;
    foreach my $triple (@$view) {
	push (@ret, $triple); # if ($triple->getAttribution->supplied);
    }
    return @ret;
}

#####
# triplesMatching (lookFor) - returns all triplets that match any of the triples in @$lookFors
# ex:	my @resourceList = $rdfDB->triplesMatching(undef, [['http://www.w3.org/schema/certHTMLv1/hasAccessTo', $node, undef]]);

sub triplesMatching {
    my ($self, $view, $lookFors, $flags) = @_;
    if (defined $view) {
	return $self->SUPER::allPredicates($view, $lookFors);
    } else {
	my @specifiers;
	foreach my $lookFor (@$lookFors) {
	    my ($p, $s, $o, $c, $r, $a) = @$lookFor;
	    my @statementQuery;
	    push (@statementQuery, 'predicate='.$p->_getId) if ($p);
	    push (@statementQuery, 'subject='.$s->_getId) if ($s);
	    push (@statementQuery, 'object='.$o->_getId) if ($o);
	    push (@statementQuery, 'containerId='.$c->_getId) if ($c);
	    push (@statementQuery, 'reifiedId='.$r->_getId) if ($r);
	    push (@statementQuery, 'attribution='.$a->_getId) if ($a);
	    push (@specifiers, '('.join (' AND ', @statementQuery).')') if (@statementQuery > 0);
	}
	my (@tables, @wheres);
	push (@tables, 'Statements');
	push (@wheres, '('.join (' OR ', @specifiers).')') if (@specifiers);
	if (exists $flags->{-attributions}) {
	    return () if (!@{$flags->{-attributions}});
	    push (@tables, 'AttributionClosures');
	    push (@wheres, 'Statements.attribution=AttributionClosures.sub ');
	    push (@wheres, 'AttributionClosures.super IN ('.join (',', map {$_->_getId} @{$flags->{-attributions}}).') ');
	}
	my $query = 'SELECT Statements.id FROM '.join (',', @tables);
	$query .= ' WHERE '.join (' AND ', @wheres) if (@wheres);
	return $self->loadTable($query, 'Statements');
    }
}

#####
# anySubjects (lookFor) - returns all triplets with this subject lookFor

sub anySubjects {
    my ($self, $view, $lookFors) = @_;
    if (defined $view) {
	return $self->SUPER::allPredicates($view, $lookFors);
    } else {
	$lookFors = join(',', map {$_->_getId} @$lookFors);
	my $query = 'SELECT id FROM Statements WHERE subject IN ('.$lookFors.')';
	return $self->loadTable($query, 'Statements');
    }
}

sub anyObjects {
    my ($self, $view, $lookFors) = @_;
    if (defined $view) {
	return $self->SUPER::allPredicates($view, $lookFors);
    } else {
	$lookFors = join(',', map {$_->_getId} @$lookFors);
	my $query = 'SELECT id FROM Statements WHERE object IN ('.$lookFors.')';
	return $self->loadTable($query, 'Statements');
    }
}

# return the list of Containers that contain the elements in lookFors.
sub anyContainerNodes {
    my ($self, $view, $lookFors) = @_;
    my @ret;
    foreach my $rdfElement ($self->anyObjects($view, $lookFors)) {
	push (@ret, $rdfElement) if ($rdfElement->_getPredicate->getUri =~ m/\A \Q$self->{RDF_SCHEMA_URI}\E (?:(li) | (?:_ (\d+))) \Z/x);
    }
    return @ret;
}

sub nodesWithType {
    my ($self, $view, $lookFors, $flags) = @_;
    $lookFors = [$lookFors] unless (ref $lookFors eq 'ARRAY');
    if (defined $view) {
	return $self->SUPER::allPredicates($view, $lookFors);
    } else {
	$lookFors = join(',', map {$_->_getId} @$lookFors);
	my @tables = qw(TypeClosures);
	my @wheres = 'TypeClosures.type IN ('.$lookFors.')';
	if ($flags->{-attributions}) {
	    push (@tables, ('AttributionClosures'));
	    push (@wheres, ('TypeClosures.attribution=AttributionClosures.sub', 
			    'AttributionClosures.super IN ('.join (',', map {$_->_getId} @{$flags->{-attributions}}).')'));
	}
	my $query = 'SELECT node FROM '.join (',', @tables).' WHERE '.join (' AND ', @wheres);
	return $self->loadTable($query, 'RdfIds');
    }
}

sub allPredicates {
    my ($self, $view, $lookFors) = @_;
    if (defined $view) {
	return $self->SUPER::allPredicates($view, $lookFors);
    } else {
	my (@selects, @asz, @whereSubjects, @wherePredicates);
	for (my $i = 0; $i < @$lookFors; $i++) {
	    my $ch = chr(ord('a')+ $i);
	    push (@asz, 'Statements AS '.$ch);
	    push (@whereSubjects, $ch.'.subject=a.subject') if ($ch ne 'a');
	    push (@wherePredicates, $ch.'.predicate='.$lookFors->[$i]->_getId);
	}
	my $query = 'SELECT a.subject FROM '.join (',', @asz).' WHERE '.
	    join (' AND ', @whereSubjects).' AND '.join (' AND ', @wherePredicates).' GROUP BY a.subject';
	return $self->loadTable($query, 'RdfIds');
    }
}

sub join {
    my ($self, $view, $nodes, $lookFors) = @_;
    return () if (@$nodes == 0);
    $lookFors = [$lookFors] unless (ref $lookFors eq 'ARRAY');

    if (defined $view) {
	return $self->SUPER::allPredicates($view, $lookFors);
    } else {
	my (@selects, @asz, @whereSubjects, @wherePredicates);
	my $nodeList = join(',', map {$_->_getId} @$nodes);
	for (my $i = 0; $i < @$lookFors; $i++) {
	    my $ch = chr(ord('a')+ $i);
	    push (@selects, $ch.'.object');
	    push (@asz, 'Statements AS '.$ch);
	    push (@whereSubjects, $ch eq 'a' ? 'a.subject IN ('.$nodeList.')' : $ch.'.subject=a.subject');
	    push (@wherePredicates, $ch.'.predicate='.$lookFors->[$i]->_getId);
	}
	my @rows;
	my $count = $self->{OB}->executeArrayQuery(\@rows, 'SELECT '.join (',', @selects).' FROM '.join (',', @asz).' WHERE '.
						   join (' AND ', @whereSubjects).' AND '.join (' AND ', @wherePredicates));
	my %patch;
	for (my $r = 0; $r < @rows; $r++) {
	    for (my $c = 0; $c < @{$rows[$r]}; $c++) {
		push (@{$patch{$rows[$r][$c]}}, [$r, $c]);
	    }
	}
	my @nodeIds = keys %patch;
	$self->{OB}->load('RdfIds', \@nodeIds, \@W3C::Rdf::RdfObjects::_TableOrder);
	foreach my $nodeId (@nodeIds) {
	    my $node = $self->{OB}->getById('RdfIds', $nodeId);
	    foreach my $patch (@{$patch{$nodeId}}) {
		my ($r, $c) = @$patch;
		$rows[$r][$c] = $node;
	    }
	}
	return @rows;
    }
}

sub members {
    my ($self, $view, $bagIds) = @_;
    my @ret;
    foreach my $bag ($self->anySubjects($view, $bagIds)) {
	foreach my $member ($self->anySubjects($view, [$bag->_getObject])) {
	    push (@ret, $member);
	}
    }
    return @ret;
}

sub show {
    my ($self, $view, $prefix, $namespaceHandler, $flags) = @_;
    $prefix = '  ' if (!defined $prefix);
    $flags = {} if (!defined $flags);

    my $ret = '';
    $ret .= $prefix.$self."\n";
    $ret .= $prefix.'RDF_SCHEMA_URI: '.$self->{RDF_SCHEMA_URI}."\n";
    if ($self->{CURRENT_SOURCE_ATTRIBUTION}) {
	$ret .= $prefix.'CURRENT_SOURCE_ATTRIBUTION: '.$self->{CURRENT_SOURCE_ATTRIBUTION}->show."\n" ;
    }
    $ret .= $prefix."NAMESPACES:\n";map {$ret .= $prefix.'  '.$_.': '.$self->{NAMESPACES}{$_}."\n"} keys %{$self->{NAMESPACES}};
    $ret .= $prefix."TRIPLES:\n";
    $ret .= $self->showTriples($view, $prefix.'  ', $namespaceHandler, $flags);
    return $ret;
}

sub expandStaementsTEST {
    my ($self, $view, $prefix, $namespaceHandler, $flags) = @_;
    delete $self->{CACHE}; # for testing cache refil
    $self->{RDF_DB}->{TRIPLES} = []; # throw away old triples
    return $self->SUPER::expandStatements($view, $prefix, $namespaceHandler, $flags);
}

######### optimize Algernon query mechanism to make it crank with SQL queries #########

sub interpretAlgernon (\@) { # for Algernon
    my ($self, $structure, $baseUri) = @_;
    my $directive = shift (@$structure);
    my @messages;
    if ($directive->content eq 'ask') {
	my $pattern = shift (@$structure) ->content;
	my (@aliases, @asz, @wheres, %vars, @selects, @labels);
	%vars = ();
	for (my $i = 0; $i < @$pattern; $i++) {
	    my ($p, $s, $o, $a, $r, $b) = @{$pattern->[$i]->content};
	    push (@messages,  'p:'.$p->show.' s:'.$s->show.' o:'.$o->show);
	    my $ch = chr(ord('a')+ $i);
	    push (@aliases, $ch.'.id');
	    push (@asz, 'Statements AS '.$ch);
	    $self->parseTripleList($p, $ch.'.predicate', \@wheres, \%vars, $baseUri);
	    $self->parseTripleList($s, $ch.'.subject', \@wheres, \%vars, $baseUri);
	    $self->parseTripleList($o, $ch.'.object', \@wheres, \%vars, $baseUri);
	    if (defined $a) {
		push (@asz, 'AttributionClosures AS '.$ch.'AC');
		push (@asz, 'Attributions AS '.$ch.'A');
		push (@wheres, $ch.'.attribution='.$ch.'AC.sub');
		push (@wheres, $ch.'AC.super='.$ch.'A.id');
		$self->parseTripleList($a, $ch.'A.uri', \@wheres, \%vars, $baseUri);
	    }
	}
	foreach my $var (keys %vars) {
	    if (@{$vars{$var}} > 1) {
		push (@wheres, map {$_.'='.$vars{$var}[0]} @{$vars{$var}}[1..$#{$vars{$var}}]);
	    }
	}
	my $command = shift (@$structure) ->content;
	if ($command eq ':collect') {
	    my $atom = shift (@$structure) ->content;
	    if (ref $atom eq 'ARRAY') {
#		push (@selects, '"("');
		foreach my $atom (@$atom) {
		    if ($atom->isa(qw(W3C::SExpr::Atom::Symbol))) {
			$atom = $atom->content;
			if ($atom =~ m/^\?(.*)/) {
			    $atom = $1;
			    push (@selects, $vars{$atom}[0]);
			    push (@labels, $atom);
			} elsif ($atom =~ m/^(\w+)\:\/\/([\w\d\-]+)/) {
			    push (@labels, $atom);
			    $self->absoluteLink($baseUri, \$atom);
			    push (@selects, $atom->_getId);
			} else {
			    die 'what kind of symbol is "'.$atom.'"?';
			}
		    } elsif ($atom->isa(qw(W3C::SExpr::Atom::String))) {
		    } else {
			die "what to do with \"$atom\"?";
		    }
		}
#		push (@selects, '")"');
	    } elsif ($atom =~ m/^\?(.*)/) {
		$atom = $1;
		push (@selects, $vars{$atom}[0]);
		push (@labels, $atom);
	    } elsif ($atom =~ m/^(\w+)\:\/\/([\w\d\-]+)/) {
		push (@labels, $atom);
		$self->absoluteLink($baseUri, \$atom);
		push (@selects, $atom->_getId);
	    } else {
		die 'what kind of symbol is "'.$atom.'"?';
	    }
	} else {
	    die "this interpreter only takes ':collect' as an 'ask' command.";
	}
	my $where = @wheres > 0 ? (' WHERE '.CORE::join (' AND ', @wheres)) : '';
	my $query = 'SELECT '.CORE::join (',', @selects).','.CORE::join (',', @aliases).' FROM '.CORE::join (',', @asz).$where; # .' LIMIT 10';
	push (@messages, $query);
#	return ([$self->loadArrayTable($query, 'RdfIds')], \@labels, \@messages, @statements);
	my @retAndStatements = ($self->loadArrayTable($query, [[scalar @selects, 'RdfIds'],[scalar @aliases, 'Statements']]));
	my (@ret, @statements);
	foreach my $row (@retAndStatements) {
	    push (@ret, [@{$row}[0..@selects-1]]);
	    push (@statements, [@{$row}[@selects..@$row-1]]);
	}
	return (\@ret, \@labels, \@messages, \@statements);
    } else {
	die "this interpreter only supports 'ask'.";
    }
}

sub parseTripleList ($$\@\%$) {
    my ($self, $atom, $field, $wheres, $vars, $baseUri) = @_;
    if ($atom->isa(qw(W3C::SExpr::Atom::String))) {
	$atom = $atom->content;
	$self->absoluteString(\$atom);
	push (@$wheres, $field.'='.$atom->_getId);
    } elsif ($atom->isa(qw(W3C::SExpr::Atom::Symbol))) {
	$atom = $atom->content;
	if ($atom =~ m/^\?(.*)/) {
	    $atom = $1;
	    push (@{$vars->{$atom}}, $field);
	} elsif ($atom =~ m/^(\w+)\:(\/\/([\w\d\-]+))?/) {
	    $self->absoluteLink($baseUri, \$atom);
	    if ($field =~ m/A\.uri/) {
		push (@$wheres, $field.'='.$atom->_getUri->_getId); # !!! my change if Attributions point to RdfIds
	    } else {
		push (@$wheres, $field.'='.$atom->_getId);
	    }
	} else {
	    die 'what kind of symbol is "'.$atom.'"?';
	}
    } else {
	die 'internal error: got a "'.$atom.'"';
    }
}


1;


__END__

=head1 NAME

W3C::Rdf::MysqlDB - database to store RDF triples

=head1 SYNOPSIS

  use W3C::Rdf::MysqlDB;
  my $MysqlDB = new W3C::Rdf::MysqlDB;
  my $desc = $MysqlDB->addDescription(...);
  $desc->addPredicateObject($predicate, $object);
  $MysqlDB->showTriples;

=head1 DESCRIPTION

Database to store RDF triples and a bunch of query operations to retrieve them.

This module is part of the W3C::Rdf CPAN module.

=head1 AUTHOR

Eric Prud\'hommeaux <eric@w3.org>

=head1 SEE ALSO

W3C::Rdf::RdfParser(3) perl(1).

=head1 COPYRIGHT

Copyright Massachusetts Institute of technology, 1998.

THIS SOFTWARE AND DOCUMENTATION IS PROVIDED "AS IS," AND COPYRIGHT HOLDERS MAKE NO REPRESENTATIONS
OR WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO, WARRANTIES OF MERCHANTABILITY OR
FITNESS FOR ANY PARTICULAR PURPOSE OR THAT THE USE OF THE SOFTWARE OR DOCUMENTATION WILL NOT INFRINGE
ANY THIRD PARTY PATENTS, COPYRIGHTS, TRADEMARKS OR OTHER RIGHTS. 

COPYRIGHT HOLDERS WILL NOT BE LIABLE FOR ANY DIRECT, INDIRECT, SPECIAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF ANY USE OF THE SOFTWARE OR DOCUMENTATION. 

The name and trademarks of copyright holders may NOT be used in advertising or publicity pertaining 
to the software without specific, written prior permission. Title to copyright in this software and 
any associated documentation will at all times remain with copyright holders. 

=cut
